import string
from base64 import b64encode
from contextlib import contextmanager
from itertools import chain
from textwrap import dedent

import ldap3
import pexpect

import ocflib.constants as constants
from ocflib.misc.mail import send_problem_report


@contextmanager
def ldap_connection(host):
    """Context manager that provides an ldap3 Connection.

    Example usage:

       with ldap_connection('ldap.ocf.berkeley.edu') as c:
            c.search(OCF_LDAP_PEOPLE, '(uid=ckuehl)', attributes=['uidNumber'])

    You might find it more convenient to use the ldap_ocf or ldap_ucb functions
    also defined.

    :param host: server hostname
    """
    server = ldap3.Server(host, use_ssl=True)
    with ldap3.Connection(server) as connection:
        yield connection


def ldap_ocf():
    """Context manager that provides an ldap3 Connection to OCF's LDAP server.

    Example usage:

       with ldap_ocf() as c:
            c.search(OCF_LDAP_PEOPLE, '(uid=ckuehl)', attributes=['uidNumber'])
    """
    return ldap_connection(constants.OCF_LDAP)


def ldap_ucb():
    """Context manager that provides an ldap3 Connection to the campus LDAP.

    Example usage:

       with ldap_ucb() as c:
            c.search(UCB_LDAP_PEOPLE, '(uid=ckuehl)', attributes=['uidNumber'])
    """
    return ldap_connection(constants.UCB_LDAP)


def create_ldap_entry_with_keytab(
    dn,
    attributes,
    keytab,
    admin_principal,
):
    """Creates an LDAP entry by shelling out to ldapadd.

    :param dn: distinguished name of the new entry
    :param attributes: dict mapping attribute name to list of values
    :param keytab: path to the admin keytab
    :param admin_principal: admin principal to use with the keytab
    """
    # LDAP attributes can have multiple values, but commonly we don't consider
    # that. So, let's sanity check the types we've received.
    for v in attributes.values():
        assert type(v) in (list, tuple), 'Value must be list or tuple.'

    def format_attr(key, values):
        # might be possible to have non-ASCII letters in keys, but don't think
        # it will happen to us. we can fix this if it ever does.
        assert all(c in string.ascii_letters for c in key), 'key is not ASCII letters'

        # rather than try to carefully escape values, we just base64 encode
        return (
            '{key}:: {value}'.format(
                key=key,
                value=b64encode(value.encode('utf8')).decode('ascii'),
            ) for value in values
        )

    lines = list(chain(
        format_attr('dn', [dn]),
        *(format_attr(key, values) for key, values in attributes.items())
    ))

    cmd = 'kinit -t {keytab} {principal} ldapadd'.format(
        keytab=keytab,
        principal=admin_principal,
    )

    child = pexpect.spawn(cmd, timeout=10)
    child.expect('SASL data security layer installed.')

    for line in lines:
        child.sendline(line)

    child.sendeof()
    child.expect('adding new entry "{dn}"'.format(dn=dn))
    child.expect(pexpect.EOF)

    output_after_adding = child.before.decode('utf8').strip()

    if 'Already exists (68)' in output_after_adding:
        raise ValueError('DN already exists, this is a duplicate.')

    if output_after_adding != '':
        send_problem_report(
            dedent(
                '''\
                Unknown problem occured when trying to add LDAP entry; the code
                should be updated to handle this case.

                dn: {dn}
                keytab: {keytab}
                principal: {principal}

                Unexpected output:
                {output_after_adding}

                Lines passed to ldapadd:
                {lines}
                '''
            ).format(
                dn=dn,
                keytab=keytab,
                principal=admin_principal,
                output_after_adding=output_after_adding,
                lines='\n'.join('    ' + line for line in lines)
            )
        )
        raise ValueError('Unknown LDAP failure was encountered.')



import mock
import pytest
import requests
import yaml
from freezegun import freeze_time

from ocflib.lab.staff_hours import _load_staff_hours
from ocflib.lab.staff_hours import get_staff_hours
from ocflib.lab.staff_hours import get_staff_hours_soonest_first
from ocflib.lab.staff_hours import Hour
from ocflib.lab.staff_hours import Staffer


TEST_HOURS = """\
staff-hours:
-   day: Monday
    time: 4:10-5:00pm
    staff: [nickimp, ckuehl, lynntsai]
    cancelled: false

-   day: Tuesday
    time: 3:10pm-4:00pm
    staff: [willh]
    cancelled: true

staff-positions:
    ckuehl: Site Manager
    willh: Deputy Manager
    nickimp: General Manager
"""


@pytest.yield_fixture
def mock_disk(tmpdir):
    f = tmpdir.join('staff_hours.yaml')
    f.write(TEST_HOURS)
    with mock.patch('ocflib.lab.staff_hours.STAFF_HOURS_FILE', f.strpath):
        yield


@pytest.yield_fixture
def mock_web():
    with mock.patch('requests.get') as m:
        yield m


@pytest.yield_fixture
def mock_not_on_disk(tmpdir):
    with mock.patch(
        'ocflib.lab.staff_hours.STAFF_HOURS_FILE',
        tmpdir.join('does-not-exist').strpath
    ):
        yield


@pytest.yield_fixture
def mock_on_web(mock_web):
    mock_web.return_value.text = TEST_HOURS
    yield


class TestLoadStaffHours:

    def test_loads_from_file_if_exists(self, mock_web, mock_disk):
        assert _load_staff_hours() == yaml.safe_load(TEST_HOURS)
        assert not mock_web.called

    def test_loads_from_web_otherwise(self, mock_not_on_disk, mock_web, mock_on_web):
        assert _load_staff_hours() == yaml.safe_load(TEST_HOURS)
        assert mock_web.called


def test_get_staff_hours(mock_disk):
    assert get_staff_hours() == [
        Hour(
            day='Monday',
            time='4:10-5:00pm',
            staff=[
                Staffer(
                    user_name='nickimp',
                    real_name='Nick Impicciche',
                    position='General Manager',
                ),
                Staffer(
                    user_name='ckuehl',
                    real_name='Chris Kuehl',
                    position='Site Manager',
                ),
                Staffer(
                    user_name='lynntsai',
                    real_name='Lynn Tsai',
                    position='Staff Member',
                ),
            ],
            cancelled=False,
        ),
        Hour(
            day='Tuesday',
            time='3:10pm-4:00pm',
            staff=[
                Staffer(
                    user_name='willh',
                    real_name='William Ho',
                    position='Deputy Manager',
                )
            ],
            cancelled=True,
        ),
    ]


@pytest.mark.parametrize('time,expected', [
    ('2015-08-23', ['Monday', 'Tuesday']),  # Sunday
    ('2015-08-24', ['Monday', 'Tuesday']),  # Monday
    ('2015-08-25', ['Tuesday', 'Monday']),  # Tuesday
    ('2015-08-26', ['Monday', 'Tuesday']),  # Wednesday
])
def test_get_staff_hours_soonest_first(mock_disk, time, expected):
    with freeze_time(time):
        assert [hour.day for hour in get_staff_hours_soonest_first()] == expected


@pytest.mark.parametrize('size', [10, 100, 1000])
def test_gravatars(size):
    staffer = Staffer(
        user_name='ckuehl',
        real_name='Chris Kuehl',
        position='Site Manager',
    )
    url = staffer.gravatar(size)
    assert url.startswith('https://www.gravatar.com/avatar/b4a3363de72988194e4b9c25195c3d07?')
    assert ('s=' + str(size)) in url.split('?')[1].split('&')
    assert requests.get(url).status_code == 200



from contextlib import contextmanager

import mock
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from ocflib.account.creation import NewAccountRequest
from ocflib.account.submission import Base
from ocflib.account.submission import get_tasks
from ocflib.account.submission import NewAccountResponse
from ocflib.account.submission import StoredNewAccountRequest
from ocflib.account.submission import user_has_request_pending
from ocflib.account.submission import username_pending
from tests.account.creation_test import fake_credentials  # noqa
from tests.account.creation_test import fake_new_account_request  # noqa
from tests.account.creation_test import mock_rsa_key  # noqa


@pytest.fixture
def session(fake_credentials):
    engine = create_engine(fake_credentials.mysql_uri)
    Base.metadata.create_all(engine)
    return sessionmaker(bind=engine)()


class TestUsernamePending:

    def test_pending(self, session, fake_new_account_request):
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request, 'reason'))
        session.commit()
        assert username_pending(session, fake_new_account_request)

    def test_not_pending(self, session, fake_new_account_request):
        assert not username_pending(session, fake_new_account_request)


class TestUserHasRequestPending:

    @pytest.mark.parametrize('attrs', [
        {'is_group': True, 'calnet_uid': None, 'callink_oid': 14},
        {'is_group': False, 'calnet_uid': 14, 'callink_oid': None},
    ])
    def test_has_pending(self, session, fake_new_account_request, attrs):
        fake_new_account_request = fake_new_account_request._replace(**attrs)
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request, 'reason'))
        session.commit()
        assert user_has_request_pending(session, fake_new_account_request)

    def test_not_has_pending(self, session, fake_new_account_request):
        assert not user_has_request_pending(session, fake_new_account_request)

    def test_not_has_pending_zero_group(self, session, fake_new_account_request):
        """callink_oid=0 can create infinite accounts."""
        fake_new_account_request = fake_new_account_request._replace(
            is_group=True,
            callink_oid=0,
            calnet_uid=None,
        )
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request, 'reason'))
        session.commit()
        assert not user_has_request_pending(session, fake_new_account_request)


@pytest.fixture
def celery_app():
    sent_messages = []

    def mock_celery_task(f):
        def update_state(**kwargs):
            pass

        # wrap everything in Mock so we can track calls
        return mock.Mock(
            side_effect=f,
            delay=mock.Mock(),
            update_state=mock.Mock(side_effect=update_state),
        )

    @contextmanager
    def dispatcher():
        def send(**kwargs):
            sent_messages.append(kwargs)
        yield mock.Mock(send=send)

    return mock.Mock(
        task=mock_celery_task,
        _sent_messages=sent_messages,
        **{
            'events.default_dispatcher': dispatcher,
        }
    )


@pytest.yield_fixture
def tasks(session, celery_app, fake_credentials):
    with mock.patch('ocflib.account.submission.sessionmaker', return_value=lambda: session):
        yield get_tasks(celery_app, credentials=fake_credentials)


@pytest.fixture
def session_with_requests(session, fake_new_account_request):
    session.add(StoredNewAccountRequest.from_request(fake_new_account_request, 'reason'))
    session.add(StoredNewAccountRequest.from_request(
        fake_new_account_request._replace(user_name='other'),
        'reason',
    ))
    session.commit()
    assert len(session.query(StoredNewAccountRequest).all()) == 2
    return session


def test_approve_request(celery_app, fake_new_account_request, session_with_requests, tasks):
    tasks.approve_request(fake_new_account_request.user_name)
    assert len(session_with_requests.query(StoredNewAccountRequest).all()) == 1

    # we want to make sure we go through the same conversion process:
    # live -> stored -> live -> dict
    request = StoredNewAccountRequest.from_request(fake_new_account_request, 'reason').to_request()
    tasks.create_account.delay.assert_called_once_with(request)
    assert celery_app._sent_messages == [
        {'type': 'ocflib.account_approved', 'request': request.to_dict()}
    ]


@mock.patch('ocflib.account.submission.send_rejected_mail')
def test_reject_request(send_rejected_mail, celery_app, fake_new_account_request, session_with_requests, tasks):
    tasks.reject_request(fake_new_account_request.user_name)
    assert len(session_with_requests.query(StoredNewAccountRequest).all()) == 1

    # we want to make sure we go through the same conversion process:
    # live -> stored -> live -> dict
    request = StoredNewAccountRequest.from_request(fake_new_account_request, 'reason').to_request()
    assert celery_app._sent_messages == [
        {'type': 'ocflib.account_rejected', 'request': request.to_dict()}
    ]

    # TODO: assert real reason
    send_rejected_mail.assert_called_once_with(request, mock.ANY)


def test_get_pending_requests(session_with_requests, tasks, fake_new_account_request):
    request = fake_new_account_request
    pending_requests = tasks.get_pending_requests()
    assert set(request.to_request() for request in pending_requests) == {
        StoredNewAccountRequest.from_request(request, 'reason').to_request(),
        StoredNewAccountRequest.from_request(
            request._replace(user_name='other'),
            'reason',
        ).to_request(),
    }


@contextmanager
def mock_validate_request(errors, warnings):
    with mock.patch(
        'ocflib.account.submission.validate_request',
        return_value=(errors, warnings),
    ):
        yield


@pytest.yield_fixture
def mock_real_create_account():
    with mock.patch('ocflib.account.submission.real_create_account') as m:
        yield m


class TestValidateThenCreateAccount:

    def test_validate_no_issues(
        self,
        tasks,
        fake_new_account_request,
        mock_real_create_account,
        fake_credentials,
        celery_app,
    ):
        with mock_validate_request([], []):
            resp = tasks.validate_then_create_account(fake_new_account_request)
            tasks.create_account.delay.assert_called_once_with(fake_new_account_request)
            assert resp == tasks.create_account.delay(fake_new_account_request).id

    @pytest.mark.parametrize('handle_warnings,expected', [
        (NewAccountRequest.WARNINGS_WARN, NewAccountResponse.FLAGGED),
        (NewAccountRequest.WARNINGS_SUBMIT, NewAccountResponse.PENDING),
        (NewAccountRequest.WARNINGS_CREATE, NewAccountResponse.CREATED),
    ])
    def test_validate_with_warnings(
        self,
        tasks,
        fake_new_account_request,
        fake_credentials,
        handle_warnings,
        expected,
        celery_app,
        session,
    ):
        assert len(session.query(StoredNewAccountRequest).all()) == 0
        with mock_validate_request([], ['ok warning']):
            request = fake_new_account_request._replace(
                handle_warnings=handle_warnings,
            )
            resp = tasks.validate_then_create_account(request)

            if expected == NewAccountResponse.CREATED:
                tasks.create_account.delay.assert_called_once_with(request)
                assert resp == tasks.create_account.delay(request).id
            else:
                assert resp == NewAccountResponse(
                    status=expected,
                    errors=['ok warning'],
                )

            if expected == NewAccountResponse.PENDING:
                assert celery_app._sent_messages == [
                    {
                        'type': 'ocflib.account_submitted',
                        'request': dict(request.to_dict(), reasons=['ok warning']),
                    }
                ]
                assert len(session.query(StoredNewAccountRequest).all()) == 1
            else:
                assert len(session.query(StoredNewAccountRequest).all()) == 0

    def test_validate_with_errors(
        self,
        tasks,
        fake_new_account_request,
        fake_credentials,
    ):
        with mock_validate_request(['bad error'], ['ok warning']):
            resp = tasks.validate_then_create_account(fake_new_account_request)
            assert resp == NewAccountResponse(
                status=NewAccountResponse.REJECTED,
                errors=['bad error', 'ok warning'],
            )
            assert not tasks.create_account.delay.called


@pytest.yield_fixture
def mock_redis_locking():
    with mock.patch('redis.from_url') as m:
        yield m


class TestCreateAccount:

    def test_create_no_issues(
        self,
        tasks,
        fake_new_account_request,
        mock_real_create_account,
        fake_credentials,
        celery_app,
        mock_redis_locking,
    ):
        with mock_validate_request([], []):
            resp = tasks.create_account(fake_new_account_request)
            assert resp == NewAccountResponse(
                status=NewAccountResponse.CREATED,
                errors=[],
            )
            mock_real_create_account.assert_called_once_with(
                fake_new_account_request,
                fake_credentials,
                mock.ANY,
            )
            assert celery_app._sent_messages == [
                {'type': 'ocflib.account_created', 'request': fake_new_account_request.to_dict()}
            ]

    @mock.patch('ocflib.account.submission.send_rejected_mail')
    def test_create_with_errors(
        self,
        send_rejected_mail,
        tasks,
        fake_new_account_request,
        mock_real_create_account,
        fake_credentials,
        mock_redis_locking,
    ):
        with mock_validate_request(['bad error'], ['ok warning']):
            resp = tasks.create_account(fake_new_account_request)
            assert resp == NewAccountResponse(
                status=NewAccountResponse.REJECTED,
                errors=['bad error', 'ok warning'],
            )
            assert not mock_real_create_account.called
            assert send_rejected_mail.called


class TestStoredNewAccountRequest:

    def test_str(self, fake_new_account_request):
        assert (
            str(StoredNewAccountRequest.from_request(fake_new_account_request, 'reason')) ==
            'someuser (individual: "Some User"), because: reason'
        )


def test_change_password(tasks, fake_credentials):
    with mock.patch('ocflib.account.submission.change_password_with_keytab') as m:
        tasks.change_password('ggroup', 'hello world', comment='comment')
        m.assert_called_once_with(
            username='ggroup',
            password='hello world',
            principal=fake_credentials.kerberos_principal,
            keytab=fake_credentials.kerberos_keytab,
            comment='comment',
        )
